import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from common.utils import round_func
from common import lut
from pathlib import Path
from common import layers
from common import losses
from common.base import SRBase
from common.transferer import TRANSFERER

class SRNetBase(SRBase):
    def __init__(self):
        super(SRNetBase, self).__init__()
        self.config = None
        self.stage1_S = layers.UpscaleBlock(None)
        self._extract_pattern_S = layers.PercievePattern(receptive_field_idxes=[[0,0],[0,1],[1,0],[1,1]], center=[0,0], window_size=2)
            
    def forward(self, x, script_config=None):
        b,c,h,w = x.shape
        x = x.reshape(b*c, 1, h, w)
        x = self.stage1_S(x, self._extract_pattern_S)
        x = x.reshape(b, c, h*self.config.upscale_factor, w*self.config.upscale_factor)
        return x

class SRNet(SRNetBase):
    def __init__(self, config):
        super(SRNet, self).__init__()
        self.config = config
        self.stage1_S.stage = layers.LinearUpscaleBlockNet(hidden_dim=self.config.hidden_dim, layers_count=self.config.layers_count, upscale_factor=self.config.upscale_factor)

class SRLut(SRNetBase):
    def __init__(self, config):
        super(SRLut, self).__init__()
        self.config = config
        self.stage1_S.stage = layers.LinearUpscaleBlockLut(quantization_interval=self.config.quantization_interval, upscale_factor=self.config.upscale_factor)

TRANSFERER.register(SRNet, SRLut)


class ChebyKANBase(SRBase):
    def __init__(self):
        super(ChebyKANBase, self).__init__()
        self.config = None
        self.stage1_S = layers.UpscaleBlock(None)
        window_size = 7
        self._extract_pattern = layers.PercievePattern(
            receptive_field_idxes=[[i,j] for i in range(window_size) for j in range(window_size)], 
            center=[window_size//2,window_size//2], 
            window_size=window_size
        )
            
    def forward(self, x, script_config=None):
        b,c,h,w = x.shape
        x = x.reshape(b*c, 1, h, w)
        x = self.stage1_S(x, self._extract_pattern)
        x = x.reshape(b, c, h*self.config.upscale_factor, w*self.config.upscale_factor)
        return x

class ChebyKANNet(ChebyKANBase):
    def __init__(self, config):
        super(ChebyKANNet, self).__init__()
        self.config = config
        window_size = 7
        self.stage1_S.stage = layers.ChebyKANUpscaleBlockNet(
            in_features=window_size*window_size, 
            out_channels=1, 
            hidden_dim=16, 
            layers_count=self.config.layers_count, 
            upscale_factor=self.config.upscale_factor, 
            degree=8
        )

class ChebyKANLut(ChebyKANBase):
    def __init__(self, config):
        super(ChebyKANLut, self).__init__()
        self.config = config
        window_size = 7
        self.stage1_S.stage = layers.ChebyKANUpscaleBlockNet(
            in_features=window_size*window_size, 
            out_channels=1, 
            hidden_dim=16, 
            layers_count=self.config.layers_count, 
            upscale_factor=self.config.upscale_factor
        ).get_lut_model(quantization_interval=self.config.quantization_interval)

    def get_loss_fn(self):
        ssim_loss = losses.SSIM(data_range=255)
        l1_loss = losses.CharbonnierLoss()
        def loss_fn(pred, target):
            return ssim_loss(pred, target) + l1_loss(pred, target) 
        return loss_fn

TRANSFERER.register(ChebyKANNet, ChebyKANLut)